"""templatetype_description

Revision ID: 5e381835e382
Revises: 52d59176f0da
Create Date: 2018-12-28 16:49:49.192505

"""
from alembic import op
import sqlalchemy as sa


import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '5e381835e382'
down_revision = '52d59176f0da'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.alter_column('tTemplateType', 'name', existing_type=sa.String(60), type_=sa.String(200), nullable=False, unique=True)
            op.add_column('tTemplateType', sa.Column('description', sa.String(1000)))
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        try:
            op.drop_table('tTemplateType_new')
        except:
            log.info("tTemplateType_new isn't there")

        try:
            op.create_table(
                'tTemplateType_new',
                sa.Column('id', sa.Integer(), primary_key=True, nullable=False),
                sa.Column('name', sa.String(200),  nullable=False, unique=True),
                sa.Column('description', sa.String(1000)),
                sa.Column('template_id', sa.Integer(), sa.ForeignKey('tTemplate.id'), nullable=False),
                sa.Column('resource_type', sa.String(200)),
                sa.Column('alias' ,sa.String(100)),

                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('layout', sa.Text(),  nullable=True)
            )

            op.execute("insert into tTemplateType_new (id, name, description, template_id, resource_type, alias, cr_date, layout) select id, name, '', template_id, resource_type, alias, cr_date, layout from tTemplate")

            op.rename_table('tTemplateType','tTemplateType_old')
            op.rename_table('tTemplateType_new', 'tTemplateType')
            op.drop_table('tTemplateType_old')

        except Exception as e:
            log.exception(e)

def downgrade():
    if op.get_bind().dialect.name == 'mysql':
        try:
            op.alter_column('tTemplateType', 'name', existing_type=sa.String(60), type_=sa.String(200), nullable=False, unique=True)
            op.drop_column('tTemplateType',  'description')
        except Exception as e:
            log.exception(e)

    else: ## sqlite
        try:
            op.create_table(
                'tTemplateType_new',
                sa.Column('id', sa.Integer(), primary_key=True, nullable=False),
                sa.Column('name', sa.String(60),  nullable=False, unique=True),
                sa.Column('template_id', sa.Integer(), sa.ForeignKey('tTemplate.id'), nullable=False),
                sa.Column('resource_type', sa.String(200)),
                sa.Column('alias' ,sa.String(100)),

                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('layout', sa.Text(),  nullable=True)
            )

            op.execute("insert into tTemplateType_new (id, name, template_id, resource_type, alias, cr_date, layout) select id, name, template_id, resource_type, alias, cr_date, layout from tTemplateType")

            op.rename_table('tTemplateType','tTemplateType_old')
            op.rename_table('tTemplateType_new', 'tTemplateType')
            op.drop_table('tTemplateType_old')

        except Exception as e:
            log.exception(e)



"""rules

Revision ID: 35088e32c557
Revises: d759891ddbf4
Create Date: 2019-11-18 14:20:23.661254

"""
from alembic import op
import sqlalchemy as sa

import datetime

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '35088e32c557'
down_revision = 'd759891ddbf4'
branch_labels = None
depends_on = None

def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.add_column('tRule', sa.Column('format', sa.String(80), nullable=True, server_default='text'))
            op.add_column('tRule', sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id'), nullable=True))
            op.add_column('tRule', sa.Column('updated_by', sa.Integer(), sa.ForeignKey('tUser.id'), nullable=True))
            op.add_column('tRule', sa.Column('updated_at', sa.DateTime(), default=datetime.datetime.utcnow(), onupdate=datetime.datetime.utcnow(), nullable=True))
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        try:
            op.drop_table('tRule_new')
        except:
            log.info('tRule_new does not exist')

        try:
            op.create_table(
                'tRule_new',
                sa.Column('id', sa.Integer(), primary_key=True, nullable=False),

                sa.Column('name', sa.String(200), nullable=False),
                sa.Column('description', sa.String(1000), nullable=True),

                sa.Column('format', sa.String(80), nullable=False, server_default='text'),

                sa.Column('ref_key', sa.String(60),  nullable=False, index=True),

                sa.Column('value', sa.Text().with_variant(sa.mysql.LONGTEXT, 'mysql'),  nullable=True),

                sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                sa.Column('scenario_id', sa.Integer(), sa.ForeignKey('tScenario.id'),  nullable=True),

                sa.Column('network_id', sa.Integer(),  sa.ForeignKey('tNetwork.id'), index=True, nullable=True),
                sa.Column('node_id', sa.Integer(),  sa.ForeignKey('tNode.id'), index=True, nullable=True),
                sa.Column('link_id', sa.Integer(),  sa.ForeignKey('tLink.id'), index=True, nullable=True),
                sa.Column('group_id', sa.Integer(),  sa.ForeignKey('tResourceGroup.id'), index=True, nullable=True),

                sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id')),
                sa.Column('updated_by', sa.Integer(), sa.ForeignKey('tUser.id')),
                sa.Column('updated_at', sa.DateTime(), sa.ForeignKey('tUser.id')),
            )

            op.execute("insert into tRule_new (id, name, description, ref_key, value, status, scenario_id, network_id, node_id, link_id, group_id, created_by, updated_by, updated_at) select id, name, description, ref_key, value, status, scenario_id, network_id, node_id, link_id, group_id, null, null, null from tRule")

            op.rename_table('tRule','tRule_old')
            op.rename_table('tRule_new', 'tRule')
            op.drop_table('tRule_old')

        except Exception as e:
            log.exception(e)


def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.drop_column('tRule', 'format')
            op.drop_column('tRule', 'created_by')
            op.drop_column('tRule', 'updated_by')
            op.drop_column('tRule', 'updated_at')
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        try:
            op.drop_table('tRule_new')
        except:
            log.info('tRule_new does not exist')

        try:
            op.create_table(
                'tRule_new',
                sa.Column('id', sa.Integer(), primary_key=True, nullable=False),

                sa.Column('name', sa.String(200), nullable=False),
                sa.Column('description', sa.String(1000), nullable=True),

                sa.Column('ref_key', sa.String(60),  nullable=False, index=True),

                sa.Column('value', sa.Text().with_variant(sa.mysql.LONGTEXT, 'mysql'),  nullable=True),

                sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                sa.Column('scenario_id', sa.Integer(), sa.ForeignKey('tScenario.id'),  nullable=True),

                sa.Column('network_id', sa.Integer(),  sa.ForeignKey('tNetwork.id'), index=True, nullable=True),
                sa.Column('node_id', sa.Integer(),  sa.ForeignKey('tNode.id'), index=True, nullable=True),
                sa.Column('link_id', sa.Integer(),  sa.ForeignKey('tLink.id'), index=True, nullable=True),
                sa.Column('group_id', sa.Integer(),  sa.ForeignKey('tResourceGroup.id'), index=True, nullable=True),

            )

            op.execute("insert into tRule_new (id, name, description, ref_key, value, status, scenario_id, network_id, node_id, link_id, group_id) select id, name, description, ref_key, value, status, scenario_id, network_id, node_id, link_id, group_id from tRule")

            op.rename_table('tRule','tRule_old')
            op.rename_table('tRule_new', 'tRule')
            op.drop_table('tRule_old')

        except Exception as e:
            log.exception(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from .base import *

from hydra_base.lib.storage import (
    MongoDatasetManager,
    MongoStorageAdapter
)

#***************************************************
# Classes definition
#***************************************************

mongo_config = MongoStorageAdapter.get_mongo_config()
mongo_storage_location_key = mongo_config["value_location_key"]
mongo_external = mongo_config["direct_location_token"]

__all__ = ['Dataset', 'DatasetCollection', 'DatasetCollectionItem', 'Metadata']

from .ownership import DatasetOwner

class Dataset(Base, Inspect, PermissionControlled, AuditMixin):
    """
        Table holding all the attribute values
    """
    __tablename__='tDataset'

    __ownerclass__ = DatasetOwner
    __ownerfk__    = 'dataset_id'

    id         = Column(Integer(), primary_key=True, index=True, nullable=False)
    name       = Column(String(200),  nullable=False)
    type       = Column(String(60),  nullable=False)
    unit_id    = Column(Integer(), ForeignKey('tUnit.id'),  nullable=True)
    hash       = Column(BIGINT(),  nullable=False, unique=True)
    cr_date    = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    hidden     = Column(String(1),  nullable=False, server_default=text(u"'N'"))
    value_ref  = Column('value', Text().with_variant(mysql.LONGTEXT, 'mysql'),  nullable=True)

    _value = MongoDatasetManager()

    @hybrid_property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        self._value = val

    @value.expression
    def value(cls):
        return cls.value_ref

    unit = relationship('Unit', backref=backref("dataset_unit", order_by=unit_id))

    _parents  = ['tResourceScenario', 'tUnit']
    _children = ['tMetadata']

    def get_value(self):
        """
            Get the value
        """
        return self.value

    def set_metadata(self, metadata_tree):
        """
        Set the metadata on a dataset.
        Note that the `mongo_storage_location_key` (MSLK) is not
        managed by this mechanism, it is instead set and deleted
        in the Dataset.value descriptor: see comments below.

        **metadata_tree**: A dictionary of metadata key-vals.
        Transforms this dict into an array of metadata objects for
        storage in the DB.
        """
        if metadata_tree is None:
            return
        if isinstance(metadata_tree, str):
            metadata_tree = json.loads(metadata_tree)

        """
        For a currently-external dataset whose value has shrunk beneath
        the size threshold, HWI will send metadata including the MSLK
        even though this is no longer applicable.
        This is deleted in the Dataset.value.__set__ action so remove
        here to avoid unwanted recreation.
        """
        metadata_tree.pop(mongo_storage_location_key, None)

        existing_metadata = []
        for m in self.metadata:
            existing_metadata.append(m.key)
            if m.key in metadata_tree:
                if m.value != metadata_tree[m.key]:
                    m.value = metadata_tree[m.key]


        for k, v in metadata_tree.items():
            if k not in existing_metadata:
                m_i = Metadata(key=str(k),value=str(v))
                self.metadata.append(m_i)

        metadata_to_delete =  set(existing_metadata).difference(set(metadata_tree.keys()))
        """
        For a dataset which is being created on external storage for the
        first time, the metadata sent by HWI will not include the MSLK,
        but this has already been added by the Dataset.value.__set__ action.
        Discard from that set here to avoid unwanted deletion.
        """
        metadata_to_delete.discard(mongo_storage_location_key)
        for m in self.metadata:
            if m.key in metadata_to_delete:
                get_session().delete(m)

    def get_val(self, timestamp=None):
        """
            If a timestamp is passed to this function,
            return the values appropriate to the requested times.

            If the timestamp is *before* the start of the timeseries data, return None
            If the timestamp is *after* the end of the timeseries data, return the last
            value.

            The raw flag indicates whether timeseries should be returned raw -- exactly
            as they are in the DB (a timeseries being a list of timeseries data objects,
            for example) or as a single python dictionary
        """
        val = get_val(self, timestamp)
        return val

    def set_hash(self,metadata=None):

        if metadata is None:
            metadata = self.get_metadata_as_dict()

        dataset_dict = {'unit_id'   : self.unit_id,
                        'type'      : self.type,
                        'value'     : self.value_ref,
                        'metadata'  : metadata}

        data_hash = generate_data_hash(dataset_dict)

        self.hash = data_hash

        return data_hash

    def get_metadata_as_dict(self):
        metadata = {}
        sortedmeta = sorted(self.metadata, key=lambda x:x.key.lower())
        for r in sortedmeta:
            val = str(r.value)

            metadata[str(r.key)] = val

        return metadata


    def _is_open(self):
        """
            Check if this dataset is globally open (the default). This negates the
            need to check for ownership
        """
        if self.hidden == 'N':
            return True

        return False


    def is_external(self):
        """
        Does the metadata indicate that this Dataset is stored in external storage?
        """
        for datum in self.metadata:
            if datum.key == mongo_storage_location_key and datum.value == mongo_external:
                return True

        return False


class DatasetCollection(Base, Inspect):
    """
    """

    __tablename__='tDatasetCollection'

    id = Column(Integer(), primary_key=True, nullable=False)
    name = Column(String(200),  nullable=False)
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    _parents  = ['tDataset']
    _children = ['tDatasetCollectionItem']

class DatasetCollectionItem(Base, Inspect):
    """
    """

    __tablename__='tDatasetCollectionItem'

    collection_id = Column(Integer(), ForeignKey('tDatasetCollection.id'), primary_key=True, nullable=False)
    dataset_id = Column(Integer(), ForeignKey('tDataset.id'), primary_key=True, nullable=False)
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    collection = relationship('DatasetCollection', backref=backref("items", order_by=dataset_id, cascade="all, delete-orphan"))
    dataset = relationship('Dataset', backref=backref("collectionitems", order_by=dataset_id,  cascade="all, delete-orphan"))

    _parents  = ['tDatasetCollection']
    _children = []

class Metadata(Base, Inspect):
    """
    """

    __tablename__='tMetadata'

    dataset_id = Column(Integer(), ForeignKey('tDataset.id', ondelete='CASCADE'), primary_key=True, nullable=False, index=True)
    key = Column(String(60), primary_key=True, nullable=False)
    value = Column(String(1000), nullable=False)

    dataset = relationship('Dataset', backref=backref("metadata", order_by=dataset_id, cascade="all, delete-orphan"))

    _parents  = ['tDataset']
    _children = []



"""Add tUser.failed_logins

Revision ID: 50bd5bac8701
Revises: 35088e32c557
Create Date: 2020-08-10 11:56:48.925174

"""
from alembic import op
import sqlalchemy as sa

import logging
log = logging.getLogger(__name__)


# revision identifiers, used by Alembic.
revision = '50bd5bac8701'
down_revision = '35088e32c557'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.add_column('tUser', sa.Column('failed_logins', sa.types.SMALLINT, nullable=True))
        except Exception as e:
            log.exception(e)


def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.drop_column('tUser', 'failed_logins')
        except Exception as e:
            log.exception(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import pytest
import hydra_base as hb
from hydra_base.lib.objects import JSONObject
from hydra_base.exceptions import HydraError


import logging
log = logging.getLogger(__name__)

class TestGroup:
    """
        Test for network-based functionality
    """
    def test_get_resourcegroup(self, client, network_with_data):
        network = network_with_data
        group = network.resourcegroups[0]
        s = network.scenarios[0]

        resourcegroup_without_data = client.get_resourcegroup(group.id)

        for ra in resourcegroup_without_data.attributes:
            assert not hasattr(ra, 'resourcescenario') or ra.resourcescenario is None

        resourcegroup_with_data = client.get_resourcegroup(group.id, s.id)

        attrs_with_data = []
        for ra in resourcegroup_with_data.attributes:
            if hasattr(ra, 'resourcescenario') and ra.resourcescenario is not None:
                if ra.resourcescenario:
                    attrs_with_data.append(ra.id)
        assert len(attrs_with_data) > 0

        group_items = client.get_resourcegroupitems(group.id, s.id)
        assert len(group_items) > 0

    def test_add_resourcegroup(self, client, network_with_data):

        network = network_with_data

        group = JSONObject({})
        group.network_id=network.id
        group.id = -1
        group.name = 'test new group'
        group.description = 'test new group'

        template_id = network.types[0].template_id
        template = JSONObject(client.get_template(template_id))

        type_summary_arr = []

        type_summary      = JSONObject({})
        type_summary.id   = template.id
        type_summary.name = template.name
        type_summary.id   = template.templatetypes[2].id
        type_summary.name = template.templatetypes[2].name

        type_summary_arr.append(type_summary)

        group.types = type_summary_arr

        new_group = client.add_group(network.id, group)

        group_attr_ids = []
        for resource_attr in new_group.attributes:
            group_attr_ids.append(resource_attr.attr_id)

        for typeattr in template.templatetypes[2].typeattrs:
            assert typeattr.attr_id in group_attr_ids

        new_network = client.get_network(network.id)

        assert len(new_network.resourcegroups) == len(network.resourcegroups)+1; "new resource group was not added correctly"

    def test_add_resourcegroupitem(self, client, network_with_extra_group):

        network = network_with_extra_group

        scenario = network.scenarios[0]

        group    = network.resourcegroups[-1]
        node_id = network.nodes[0].id

        item = JSONObject({})
        item.ref_key = 'NODE'
        item.ref_id  = node_id
        item.group_id = group.id

        new_item = client.add_resourcegroupitem(item, scenario.id)

        assert new_item.node_id == node_id


    def test_set_group_status(self, client, network_with_extra_group):
        net = network_with_extra_group

        group_to_delete = net.resourcegroups[0]

        client.set_group_status(group_to_delete.id, 'X')

        updated_net = client.get_network(net.id)

        group_ids = []
        for g in updated_net.resourcegroups:
            group_ids.append(g.id)

        assert group_to_delete.id not in group_ids

        client.set_group_status(group_to_delete.id, 'A')

        updated_net = client.get_network(net.id)

        group_ids = []
        for g in updated_net.resourcegroups:
            group_ids.append(g.id)

        assert group_to_delete.id in group_ids


    def test_purge_group(self, client, network_with_extra_group):
        net = network_with_extra_group
        scenario_id = net.scenarios[0].id
        group_id_to_delete = net.resourcegroups[-1].id
        group_id_to_keep = net.resourcegroups[0].id

        group_datasets = client.get_resource_data('GROUP', group_id_to_delete, scenario_id)
        log.info("Deleting group %s", group_id_to_delete)

        client.delete_resourcegroup(group_id_to_delete, 'Y')

        updated_net = JSONObject(client.get_network(net.id, 'Y'))
        assert len(updated_net.resourcegroups) == 1
        assert updated_net.resourcegroups[0].id == group_id_to_keep

        for rs in group_datasets:
            #In these tests, all timeseries are unique to their resources,
            #so after removing the group no timeseries to which it was attached
            #should still exist.
            d = rs.dataset
            if d.type == 'timeseries':
                with pytest.raises(HydraError):
                    client.get_dataset(d.id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from .base import *

from .ownership import NetworkOwner, ProjectOwner
from .scenario import Scenario, ResourceScenario
from .permissions import User
from .network import Network, ResourceAttr
from .attributes import Attr

global project_cache_key
project_cache_key = config.get('cache', 'projectkey', 'userprojects')

__all__ = ['Project']


class Project(Base, Inspect, PermissionControlled):
    """
    """

    __tablename__='tProject'
    ref_key = 'PROJECT'

    __ownerclass__ = ProjectOwner

    __table_args__ = (
        UniqueConstraint('name', 'created_by', 'status', name="unique proj name"),
    )

    core_columns = ['name', 'description', 'status']

    attribute_data = []

    id = Column(Integer(), primary_key=True, nullable=False)
    name = Column(String(200),  nullable=False, unique=False)
    description = Column(String(1000))
    status = Column(String(1),  nullable=False, server_default=text(u"'A'"))
    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    created_by = Column(Integer(), ForeignKey('tUser.id'), nullable=False)
    appdata = Column(JSON)
    user = relationship('User', backref=backref("projects", order_by=id))

    parent_id = Column(Integer(), ForeignKey('tProject.id'), nullable=True)
    parent = relationship('Project', remote_side=[id],
        backref=backref("children", order_by=id))

    _parents  = []
    _children = ['tNetwork']

    def get_networks(self, user_id, include_deleted_networks=False):
        """
        get all the networks, scenarios and owners of all the networks in the project.
        These are in 3 different sets:
        1: Networks of which user_id is the creator
        2: Networks of which user_id has read access but is not the creator
        3: Networks which user_id can see by virtue of being an owner of the project
           In this case, user_id can see all the networks. Ownership can be directly set
           on this project, or can be set at a higher level project
        """

        log.debug("Getting networks for project %s", self.id)

        networks = []
        networks_not_creator = []

        if self.is_owner(user_id):
            #all networks in the project, as the user owns the project
            networks_not_creator = get_session().query(Network)\
                .filter(Network.project_id == self.id).all()
        else:
            #all networks created by someone else, but which this user is an owner,
            #and this user can read this network
            networks_not_creator = get_session().query(Network).join(NetworkOwner)\
                .filter(Network.project_id == self.id)\
                .filter(Network.created_by != user_id)\
                .filter(NetworkOwner.user_id == user_id)\
                .filter(NetworkOwner.view == 'Y').all()

        all_network_ids = [n.id for n in networks_not_creator]

        #for efficiency, get all the owners in 1 query and sort them by network
        all_owners = get_session().query(NetworkOwner)\
            .filter(NetworkOwner.network_id.in_(all_network_ids)).all()

        owners_by_network = defaultdict(list)
        for owner in all_owners:
            owners_by_network[owner.network_id].append(JSONObject(owner))

        project_owners = self.get_owners()

        #for efficiency, get all the scenarios in 1 query and sort them by network
        all_scenarios = get_session().query(Scenario)\
            .filter(Scenario.network_id.in_(all_network_ids)).all()

        scenarios_by_network = defaultdict(list)
        for netscenario in all_scenarios:
            scenarios_by_network[netscenario.network_id].append(netscenario)

        for net_i in networks_not_creator:

            if include_deleted_networks is False and net_i.status.lower() == 'x':
                continue

            net_j = JSONObject(net_i)
            net_j.owners = owners_by_network[net_j.id]
            owner_ids = [no.user_id for no in owners_by_network[net_j.id]]
            #include inherited owners from the project ownership for this project
            for proj_owner in project_owners:
                if proj_owner.user_id not in owner_ids:
                    proj_owner.source = f'Inherited from: {self.name} (ID:{self.id})'
                    net_j.owners.append(proj_owner)
            net_j.scenarios = scenarios_by_network[net_j.id]
            networks.append(net_j)

        log.debug("%s networks retrieved", len(networks))

        return networks

    def get_child_projects(self, user_id, include_deleted_networks=False, levels=1):
        """
        Get all the direct child projects of a given project
        i.e. all projects which have this project specified in the 'parent_id' column
        """
        log.debug("Getting child projects of project %s", self.id)

        child_projects_i = get_session().query(Project).outerjoin(ProjectOwner)\
            .filter(Project.parent_id == self.id).all()

        projects_with_access = [] # projects to which the user has access
        for child_proj_i in child_projects_i:
            if child_proj_i.check_read_permission(user_id, do_raise=False) is True:
                projects_with_access.append(child_proj_i)

        project_lookup = {p.id:p for p in child_projects_i}

        owners = get_session().query(
            User.id.label('user_id'), User.display_name,
            ProjectOwner.project_id).filter(
                User.id==ProjectOwner.user_id).filter(
                    ProjectOwner.project_id.in_([p.id for p in child_projects_i])).all()

        creators = get_session().query(User.id.label('user_id'), User.display_name).filter(User.id.in_([p.created_by for p in child_projects_i])).all()
        creator_lookup = {u.user_id:JSONObject(u)  for u in creators}

        owner_lookup = defaultdict(list)
        for p in child_projects_i:
            owner_lookup[p.id] = [creator_lookup[p.created_by]]
        for o in owners:
            if o.user_id == project_lookup[o.project_id].created_by:
                continue
            owner_lookup[o.project_id].append(JSONObject(o))

        #Get the inherited owners of the child projects
        parentowners = self.get_owners()

        child_projects = []
        for child_proj_i in projects_with_access:
            project = JSONObject(child_proj_i)
            project.owners = owner_lookup.get(project.id, [])
            owner_ids = [o.user_id for o in project.owners]
            #add any inherited owners to the child projects.
            for parentowner in parentowners:
                if parentowner.user_id not in owner_ids:
                    parentowner.source = f"Inherited from {parentowner.project_name} (ID:{parentowner.project_id})"
                    project.owners.append(parentowner)

            project.networks = child_proj_i.get_networks(
                user_id,
                include_deleted_networks=include_deleted_networks)
            if levels > 0:
                project.projects = child_proj_i.get_child_projects(
                    user_id,
                    include_deleted_networks=include_deleted_networks, levels=(levels-1))
            else:
                project.projects = []
            child_projects.append(project)

        log.debug("%s child projects retrieved", len(child_projects))

        return child_projects

    def is_owner(self, user_id):
        """Check whether this user is an owner of this project, either directly
        #or by virtue of being an owner of a higher-up project"""

        if self.check_read_permission(user_id, nav=False, do_raise=False) is True:
            return True
        p = self.parent
        if p is not None:
            return p.is_owner(user_id)

        return False

    def get_owners(self):
        """
            Get all the owners of a project, both those which are applied directly
            to this project, but also who have been granted access via a parent project
        """

        owners = [JSONObject(o) for o in self.owners]
        owner_ids = [o.user_id for o in owners]

        for o in owners:
            o.project_name = self.name
            o.type = 'PROJECT'

        parent_owners = []
        if self.parent_id is not None:
            parent_owners = list(filter(lambda x:x.user_id not in owner_ids, self.parent.get_owners()))
            for po in parent_owners:
                po.source = f'Inherited from: {po.project_name} (ID:{po.project_id})'

        return owners + parent_owners

    """
    This map should look like:
     {'UID' :
         {
             None: [P1, P2],
             'P1': [P3, P4]
         }
     }
    Where UID is the user ID and the inner keys are project IDS, and the lists are
    projects the user can see within those projects. The 'None' key at the top is for
    top-level projects.
    """
    @classmethod
    def get_cache(cls, user_id=None):
        try:
            if user_id is None:
                return cache.get(project_cache_key, {})
            else:
                return cache.get(project_cache_key, {}).get(user_id, {})
        except Exception as e:
            log.exception(e)
            err_value = cache.get(project_cache_key, {})
            if type(err_value) is dict:
                err_value = err_value.get(user_id)
            log.warning(f"Error to get project cache: {project_cache_key}, user_id={user_id}, err_value={err_value}")
            return {}

    @classmethod
    def set_cache(cls, data):
        cache.set(project_cache_key, dict(data))

    @classmethod
    def clear_cache(cls, uid):
        projectcache = cache.get(project_cache_key, {})
        if projectcache.get(uid) is not None:
            del projectcache[uid]
        cls.set_cache(projectcache)

    def get_name(self):
        return self.project_name

    @classmethod
    def build_user_cache(cls, uid):
        """
            Build the cache of projects a user has access to either by direct Ownership
            or by indirect access required for navigating to a projct to which they own
        """
        if cls.get_cache().get(uid) is not None:
            return

        project_user_cache = defaultdict(lambda: defaultdict(list))
        ##Don't load the project's networks. Load them separately, as the networks
        #must be checked individually for ownership
        projects_qry = get_session().query(Project)
        network_project_qry = get_session().query(Project.id)

        log.info("Getting projects for user %s", uid)

        network_projects_i = network_project_qry.outerjoin(Network).outerjoin(NetworkOwner).filter(
            Network.status == 'A',
            NetworkOwner.user_id == uid,
            NetworkOwner.view == 'Y').distinct().all()

        #for some reason this outputs a list of tuples.
        projects_with_network_owner = [p[0] for p in network_projects_i]

        projects_qry = projects_qry.outerjoin(ProjectOwner).filter(
            Project.status == 'A', or_(
                and_(ProjectOwner.user_id == uid, ProjectOwner.view == 'Y'),
                Project.id.in_(projects_with_network_owner)
            )
        )

        projects_qry = projects_qry.options(noload(Project.networks)).order_by('id')

        projects_i = projects_qry.all()

        parent_project_ids = []
        for p in projects_i:
            project_user_cache[uid][p.parent_id].append(p.id)
            if p.parent_id is not None:
                parent_project_ids.append(p.parent_id)

        cls._build_user_cache_up_tree(uid, parent_project_ids, project_user_cache)

        cls._build_user_cache_down_tree(uid, [p.id for p in projects_i], project_user_cache)

        cls.set_cache(project_user_cache)

    @classmethod
    def _build_user_cache_down_tree(cls, uid, project_ids, project_user_cache):

        if len(project_ids) == 0:
            return

        projects = get_session().query(Project).filter(Project.parent_id.in_(project_ids)).all()

        child_project_ids = []
        for p in projects:
            project_user_cache[uid][p.parent_id].append(p.id)
            child_project_ids.append(p.id)

        cls._build_user_cache_down_tree(uid, child_project_ids, project_user_cache)

    @classmethod
    def _build_user_cache_up_tree(cls, uid, project_ids, project_user_cache):

        if len(project_ids) == 0:
            return

        projects = get_session().query(Project).filter(Project.id.in_(project_ids)).all()

        parent_project_ids = []
        for p in projects:
            project_user_cache[uid][p.parent_id].append(p.id)
            if p.parent_id is not None:
                parent_project_ids.append(p.parent_id)

        cls._build_user_cache_up_tree(uid, parent_project_ids, project_user_cache)


    def remove_from_cache(self):
        """Remove this project from the cache of all users who can see it"""
        projectcache = cache.get(project_cache_key, {})
        for uid, user_projects in projectcache.items():
            for parent_id, child_projects in user_projects.items():
                if self.id in child_projects:
                    child_projects.remove(self.id)
        cache.set(project_cache_key, projectcache)

    def get_attribute_data(self):
        attribute_data_rs = get_session().query(ResourceScenario).join(ResourceAttr).filter(
            ResourceAttr.project_id==self.id).all()
        #lazy load datasets
        [rs.dataset.metadata for rs in attribute_data_rs]
        [rs.resourceattr.attr for rs in attribute_data_rs]
        self.attribute_data = attribute_data_rs
        return attribute_data_rs

    def add_attribute(self, attr_id, attr_is_var='N'):
        res_attr = ResourceAttr()
        res_attr.attr_id = attr_id
        res_attr.attr_is_var = attr_is_var
        res_attr.ref_key = self.ref_key
        res_attr.project_id  = self.id
        self.attributes.append(res_attr)

        return res_attr

    def set_owner(self, user_id, read='Y', write='Y', share='Y'):

        for o in self.owners:
            if user_id == o.user_id:
                owner = o
                break
        else:
            owner = ProjectOwner()
            owner.project_id = self.id
            owner.user_id = int(user_id)
            self.owners.append(owner)

        owner.view = read
        owner.edit = write
        owner.share = share

        Project.clear_cache(user_id)

        return owner

    def unset_owner(self, user_id):
        owner = None
        if str(user_id) == str(self.created_by):
            log.warning("Cannot unset %s as owner, as they created the project", user_id)
            return
        for o in self.owners:
            if user_id == o.user_id:
                owner = o
                get_session().delete(owner)
                break
        Project.clear_cache(user_id)

    def check_read_permission(self, user_id, do_raise=True, is_admin=None, nav=True):
        """
            Check whether this user can read this project
            nav: Indicates whether you can return true if the user is not an owner but
            is allowed to see the project for navigation purposes.
        """

        has_permission = super(Project, self).check_read_permission(user_id,
            is_admin=is_admin, do_raise=False)

        if nav is True:
            Project.build_user_cache(user_id)
            for v in Project.get_cache(user_id).values():
                if self.id in v:
                    return True

        if has_permission is False and self.parent_id is not None:
            """
                Permission check up the tree only applies to non 'nav'. i.e. i a parent which
                can be accessed for nav purposes does not count.
            """
            has_permission = self.parent.check_read_permission(user_id, nav=False, do_raise=False)

        if has_permission is False and do_raise is True:
            raise PermissionError("Permission denied. User %s does not have read"
                                  " access on project '%s'" % (user_id, self.name))

        return has_permission

    def check_write_permission(self, user_id, do_raise=True, is_admin=None):
        """
            Check whether this user can write this project
        """
        has_permission = super(Project, self).check_write_permission(user_id,
            is_admin=is_admin, do_raise=False)

        if has_permission is False and self.parent_id is not None:
            has_permission = self.parent.check_write_permission(user_id)

        if has_permission is False and do_raise is True:
            raise PermissionError("Permission denied. User %s does not have edit"
                                  " access on project %s" % (user_id, self.id))

        return has_permission

    def get_scoped_attributes(self, include_hierarchy=False, name_match=None, return_json=True):
        """
            Get all the attributes scoped to this project, and to all projects above
            it in the project hierarchy (including global attributes if requested)
            args:
                include_hierarchy (Bool): Include attribtues from projects higher up in the
                    project hierarchy
        """

        scoped_attrs_qry = get_session().query(Attr).filter(Attr.project_id==self.id)

        if name_match is not None:
            name_match = name_match.lower()
            scoped_attrs_qry = scoped_attrs_qry.filter(
                func.lower(Attr.name).like(f'%{name_match}%'))

        scoped_attrs = scoped_attrs_qry.all()

        if self.parent_id is not None and include_hierarchy is True:
            scoped_attrs.extend(self.parent.get_scoped_attributes(
                include_hierarchy=True, name_match=name_match))

        if return_json is True:
            scoped_attrs_j = [JSONObject(a) for a in scoped_attrs]
            #This is for convenience to avoid having to do extra calls to get the project name
            for a in scoped_attrs_j:
                a.project_name = self.name

            return scoped_attrs_j
        else:
            return scoped_attrs

    def get_hierarchy(self, user_id):

        project_hierarchy = [JSONObject(self)]
        if self.parent_id:
            project_hierarchy = project_hierarchy + self.parent.get_hierarchy(user_id)

        return project_hierarchy



"""simplify network

Revision ID: 1eb51c6159ac
Revises: 7d6099524c02
Create Date: 2017-12-29 11:36:35.323359

"""
from alembic import op
import sqlalchemy as sa
from hydra_base import db

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '1eb51c6159ac'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###


    if op.get_bind().dialect.name == 'mysql':
        try:
            op.alter_column('tNetwork', 'network_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tNetwork', 'network_name', new_column_name='name', existing_type=sa.String(200))
            op.alter_column('tNetwork', 'network_description', new_column_name='description', existing_type=sa.String(1000))
        except Exception as e:
            log.exception(e)
    else:
        try:
            op.drop_table('tNetwork_new')
        except:
            log.info("tNetwork_new does not exist")

        try:
            op.create_table(
                'tNetwork_new',
                sa.Column('id', sa.Integer, primary_key=True),
                sa.Column('name', sa.Text(200), nullable=False),
                sa.Column('description', sa.Text(1000)),
                sa.Column('layout', sa.Text(1000)),
                sa.Column('project_id', sa.Integer(), sa.ForeignKey('tProject.project_id'),  nullable=False),
                sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('projection', sa.String(1000)),
                sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id')),
                sa.UniqueConstraint('name', 'project_id', name="unique net name")
            )

            op.execute("insert into tNetwork_new (id, name, description, layout, project_id, status, cr_date, projection, created_by) select network_id, network_name, network_description, layout, project_id, status, cr_date, projection, created_by from tNetwork")


            op.rename_table('tNetwork','tNetwork_old')
            op.rename_table('tNetwork_new', 'tNetwork')
            op.drop_table('tNetwork_old')
        except Exception as e:
            log.exception(e)

    # ### end Alembic commands ###


def downgrade():

    if op.get_bind().dialect.name == 'mysql':
        try:
            op.alter_column('tNetwork', 'network_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tNetwork', 'network_name', new_column_name='name', existing_type=sa.String(200))
            op.alter_column('tNetwork', 'network_description', new_column_name='description', existing_type=sa.String(1000))
        except Exception as e:
            log.exception(e)
    else:
        try:
            op.drop_table('tNetwork_new')
        except:
            log.info("tNetwork_new does not exist")

        try:
            op.create_table(
                'tNetwork_new',
                sa.Column('network_id', sa.Integer, primary_key=True),
                sa.Column('network_name', sa.Text(200), nullable=False),
                sa.Column('network_description', sa.Text(1000)),
                sa.Column('layout', sa.Text(1000)),
                sa.Column('project_id', sa.Integer(), sa.ForeignKey('tProject.project_id'),  nullable=False),
                sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('projection', sa.String(1000)),
                sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id')),
                sa.UniqueConstraint('name', 'project_id', name="unique net name")
            )

            op.execute("insert into tNetwork_new (network_id, network_name, network_description, layout, project_id, status, cr_date, projection, created_by) select id, name, description, layout, project_id, status, cr_date, projection, created_by from tNetwork")


            op.rename_table('tNetwork','tNetwork_old')
            op.rename_table('tNetwork_new', 'tNetwork')
            op.drop_table('tNetwork_old')
        except Exception as e:
            log.exception(e)



"""
  Types that can be represented by a dataset are defined here

  Each Hydra type must subclass DataType and implement the
  required abstract properties and methods.  The form of each
  class' constructor is not part of the interface and is left
  to the implementer.
"""
import json
import math
import six
import numpy as np
import pandas as pd
from io import StringIO
from abc import abstractmethod, abstractproperty
from datetime import datetime
import collections
from hydra_base import config

from .Encodings import ScalarJSON, ArrayJSON, DescriptorJSON, DataframeJSON, TimeseriesJSON
from hydra_base.exceptions import HydraError

import logging
log = logging.getLogger(__name__)


class DataType(object):
    """ The DataType class serves as an abstract base class for data types"""
    def __init_subclass__(cls):
        tag = cls.tag
        name = cls.name

        # Register class with hydra
        from .Registry import typemap
        if tag in typemap:
            raise ValueError('Type with tag "{}" already registered.'.format(tag))
        else:
            typemap[tag] = cls
            log.info('Registering data type "{}".'.format(tag))



    @abstractproperty
    def skeleton(self):
        """ Reserved for future use """
        pass

    @abstractproperty
    def tag(self):
        """ A str which uniquely identifies this type and serves as its key in
            the Registry.typemap dict
        """
        pass

    @abstractproperty
    def value(self):
        """ This type's representation of the value contained within
            a dataset of the same type
        """
        pass

    @abstractmethod
    def validate(self):
        """ Raises (any) exception if the dataset's value argument
            cannot be correctly represented as this type
        """
        pass

    @abstractmethod
    def json(self):
        """ Reserved for future use """
        pass

    @abstractmethod
    def fromDataset(cls, value, metadata=None):
        """ Factory method which performs any required transformations
            on a dataset argument, invokes the type's ctor, and returns
            the resulting instance
        """
        pass


class Scalar(DataType):
    tag      = "SCALAR"
    name     = "Scalar"
    skeleton = "[%f]"
    json     = ScalarJSON()

    def __init__(self, value):
        super(Scalar, self).__init__()
        self.value = value
        self.validate()

    @classmethod
    def fromDataset(cls, value, metadata=None):
        return cls(value)

    def validate(self):
        f = float(self.value)

    def get_value(self):
        return str(self._value)

    def set_value(self, val):
        self._value = val

    value = property(get_value, set_value)


class Array(DataType):
    tag      = "ARRAY"
    name     = "Array"
    skeleton = "[%f, ...]"
    json     = ArrayJSON()

    def __init__(self, encstr):
        super(Array, self).__init__()
        self.value = encstr
        self.validate()

    @classmethod
    def fromDataset(cls, value, metadata=None):
        return cls(value)

    def validate(self):
        j = json.loads(self.value)
        assert len(j) > 0           # Sized
        assert iter(j) is not None  # Iterable
        assert j.__getitem__        # Container
        assert not isinstance(j, six.string_types) # Exclude strs

    def get_value(self):
        return self._value

    def set_value(self, val):
        self._value = val

    value = property(get_value, set_value)


class Descriptor(DataType):
    tag      = "DESCRIPTOR"
    name     = "Descriptor"
    skeleton = "%s"
    json     = DescriptorJSON()

    def __init__(self, data):
        super(Descriptor, self).__init__()
        self.value = data
        self.validate()

    @classmethod
    def fromDataset(cls, value, metadata=None):
        if metadata and metadata.get('data_type') == 'hashtable':
            try:
                df = pd.read_json(six.text_type(value))
                data = df.transpose().to_json()
            except Exception:
                noindexdata = json.loads(six.text_type(value))
                indexeddata = {0:noindexdata}
                data = json.dumps(indexeddata)
            return cls(data)
        else:
            return cls(six.text_type(value))


    def validate(self):
        pass

    def get_value(self):
        return self._value

    def set_value(self, val):
        self._value = val

    value = property(get_value, set_value)


class Dataframe(DataType):
    tag      = "DATAFRAME"
    name     = "Data Frame"
    skeleton = "%s"
    json     = DataframeJSON()

    def __init__(self, data):
        super(Dataframe, self).__init__()
        self.value = data
        self.validate()

    @classmethod
    def fromDataset(cls, value, metadata=None):

        df = cls._create_dataframe(value)

        return cls(df)

    @classmethod
    def _create_dataframe(cls, value):
        """
            Builds a dataframe from the value
        """
        try:

            ordered_jo = json.loads(six.text_type(value), object_pairs_hook=collections.OrderedDict)

            #Pandas does not maintain the order of dicts, so we must break the dict
            #up and put it into the dataframe manually to maintain the order.

            cols = list(ordered_jo.keys())

            if len(cols) == 0:
                raise ValueError("Dataframe has no columns")

            #Assume all sub-dicts have the same set of keys
            if isinstance(ordered_jo[cols[0]], list):
                index = range(len(ordered_jo[cols[0]]))
            else:
                #cater for when the indices are not the same by identifying
                #all the indices, and then making a set of them.
                longest_index = []
                for col in ordered_jo.keys():
                    index = list(ordered_jo[col].keys())
                    if len(index) > len(longest_index):
                        longest_index = index
                index = longest_index

            df = pd.read_json(StringIO(value), convert_axes=False)

            #Make both indices the same type, so they can be compared
            df.index = df.index.astype(str)
            new_index = pd.Index(index).astype(str)

            #Now reindex the dataframe so that the index is in the correct order,
            #as per the data in the DB, and not with the default pandas ordering.
            new_df = df.reindex(new_index)

            #If the reindex didn't work, don't use that value
            if new_df.isnull().sum().sum() != len(df.index):
                df = new_df


        except ValueError as e:
            """ Raised on scalar types used as pd.DataFrame values
                in absence of index arg
            """
            log.exception(e)
            raise HydraError(str(e))

        except AssertionError as e:
            log.warning("An error occurred creating the new data frame: %s. Defaulting to a simple read_json"%(e))
            df = pd.read_json(value).fillna(0)

        return df

    def validate(self):
        assert isinstance(self._value, pd.DataFrame)
        assert not self._value.empty


    def get_value(self):
        return self._value.to_json()

    def set_value(self, val):
        self._value = val
        try:
            """ Use validate test to confirm is pd.DataFrame... """
            self.validate()
        except AssertionError:
            """ ...otherwise attempt as json..."""
            try:
                df = self.__class__._create_dataframe(val)
                self._value = df
                self.validate()
            except Exception as e:
                """ ...and fail if neither """
                raise HydraError(str(e))

    value = property(get_value, set_value)

class Timeseries(DataType):
    tag      = "TIMESERIES"
    name     = "Time Series"
    skeleton = "[%s, ...]"
    json     = TimeseriesJSON()

    def __init__(self, ts):
        super(Timeseries, self).__init__()
        self.value = ts
        self.validate()

    @classmethod
    def fromDataset(cls, value, metadata=None):
        ordered_jo = json.loads(six.text_type(value), object_pairs_hook=collections.OrderedDict)
        ts = pd.DataFrame.from_dict(ordered_jo)
        return cls(ts)


    def validate(self):
        base_ts = pd.Timestamp("01-01-1970")
        #TODO: We need a more permanent solution to seasonal/repeating timeseries
        seasonal_year = config.get('DEFAULT','seasonal_year', '1678')
        seasonal_key = config.get('DEFAULT', 'seasonal_key', '9999')
        jd = json.loads(self.value, object_pairs_hook=collections.OrderedDict)
        for k,v in jd.items():
            for date in (six.text_type(d) for d in v.keys()):
                #A date with '9999' in it is a special case, but is an invalid year
                #for pandas, so replace it with the first year allowed by pandas -- 1678
                if date.find(seasonal_key) >= 0:
                    date = date.replace(seasonal_key, seasonal_year)

                ts = pd.Timestamp(date)

                assert isinstance(ts, base_ts.__class__) # Same type as known valid ts


    def get_value(self):
        return self._value.to_json(date_format='iso', date_unit='ns')

    def set_value(self, val):
        self._value = val

    value = property(get_value, set_value)



"""attribute_scoping

Revision ID: a81a860cda39
Revises: cec2b77ad85e
Create Date: 2021-12-16 13:48:55.286610

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'a81a860cda39'
down_revision = 'cec2b77ad85e'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.add_column('tAttr', sa.Column('project_id', sa.Integer(), sa.ForeignKey('tProject.id'), nullable=True))
        except Exception as e:
            log.critical(e)

        try:
            op.add_column('tAttr', sa.Column('network_id', sa.Integer(), sa.ForeignKey('tNetwork.id'), nullable=True))
        except Exception as e:
            log.critical(e)

def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.drop_column('tAttr', sa.Column('project_id', sa.Integer(), sa.ForeignKey('tProject.id'), nullable=True))
        except Exception as e:
            log.critical(e)

        try:
            op.drop_column('tAttr', sa.Column('network_id', sa.Integer(), sa.ForeignKey('tNetwork.id'), nullable=True))
        except Exception as e:
            log.critical(e)



import json
import pandas as pd


""" Descriptor for Scalar JSON encoding/decoding"""
class ScalarJSON(object):
    def __get__(self, instance, owner):
        return json.dumps({owner.tag : { arg : getattr(instance, arg) for arg in owner.ctor_fmt }})

    def __set__(self, instance, encstr):
        jo = json.loads(encstr)[instance.tag]
        for arg in instance.ctor_fmt:
            setattr(instance, arg, jo[arg])


""" Descriptor for Array JSON encoding/decoding"""
class ArrayJSON(object):
    def __get__(self, instance, owner):
        return json.dumps({owner.tag : instance.data.to_dict()})

    def __set__(self, instance, encstr):
        data = json.loads(encstr)[instance.tag]
        instance.data = pd.DataFrame.from_dict(data)


""" Descriptor for (Hydra) Descriptor JSON encoding/decoding"""
class DescriptorJSON(object):
    def __get__(self, instance, owner):
        return ""

    def __set__(self, instance, encstr):
        pass


""" Descriptor for Dataframe JSON encoding/decoding"""
class DataframeJSON(object):
    def __get__(self, instance, owner):
        return "{}"

    def __set__(self, instance, encstr):
        pass


""" Descriptor for Timeseries JSON encoding/decoding"""
class TimeseriesJSON(object):
    def __get__(self, instance, owner):
        return "{}"

    def __set__(self, instance, encstr):
        pass



from unittest.mock import patch
import numpy as np
import json
import os
import pandas as pd
import pytest

from packaging import version

from hydra_base.lib.storage import HdfStorageAdapter
from hydra_base.util import NullAdapter

min_lib_versions = {
    "fsspec": version.parse("0.9.0"),
    "h5py": version.parse("3.7.0"),
    "s3fs": version.parse("0.6.0")
}

@pytest.fixture
def hdf():
    hdf = HdfStorageAdapter()
    return hdf

@pytest.fixture
def hdf_config():
    return HdfStorageAdapter.get_hdf_config()

@pytest.fixture
def public_aws_file():
    return {
        "path": "s3://modelers-data-bucket/unit-tests/ETH_flow_sim.h5",
        "file_size": 7374454,
        "series_name": "BR_Kabura",
        "series_size": 12784,
        "series_type": "float64"
    }

IAM_ACCESS_KEY = os.getenv("TEST_AWS_ACCESS_KEY_ID")
IAM_SECRET_KEY = os.getenv("TEST_AWS_SECRET_ACCESS_KEY")

@pytest.fixture
def private_aws_file():
    return {
        "path": "s3://private-test-data/ETH_flow_sim.h5",
        "file_size": 7374454,
        "series_name": "BR_Kabura",
        "series_size": 12784,
        "series_type": "float64"
    }

@pytest.fixture
def multigroup_file():
    return {
        "path": "s3://modelers-data-bucket/unit-tests/grid_data.h5",  # NB rot13 strings
        "groups": ['RFJ_Rffrk_erfhygf', 'prageny_fbhgu_rffrk_erfhygf', 'qnvyl_cebsvyrf', 'yvapbyafuver_erfhygf', 'zbaguyl_cebsvyrf', 'gvzrfrevrf']
    }

@pytest.fixture
def non_timeseries_index_file():
    return {
        "path": "s3://modelers-data-bucket/unit-tests/synthetic_flow_duration_recorder.h5",
        "group": "hydra_test_data"
    }

@pytest.fixture(params=["s3://modelers-data-bucket/does_not_exist.h5", "does_not_exist"])
def bad_url(request):
    return request.param

@pytest.fixture
def symlink_path(hdf):
    symlink = os.path.join(hdf.filestore_path, "symlink.txt")
    os.symlink("/dev/shm/somefile.txt", symlink)
    yield symlink
    os.unlink(symlink)

bad_paths = {
    "relative_component": "../../dev/shm/somefile.txt",
    "variable_component": "$HOME/somefile.txt",
    "user_component": "~/somefile.txt",
    "device_file": "/dev/null"
}


class TestHdf():
    def test_exists(self, hdf, hdf_config):
        """
          Can the HDF adapter be instantiated, and does the config exist?
        """
        assert hdf
        assert hdf_config

    def test_lib_versions(self):
        """
          Are the required libraries present and adequate versions?
        """
        import importlib
        for libname, semver in min_lib_versions.items():
            lib = importlib.import_module(libname)
            assert version.parse(lib.__version__) >= semver

    @pytest.mark.requires_hdf
    def test_hdf_size(self, hdf, public_aws_file):
        """
          Does the reported file size match an expected value?
        """
        assert hdf.file_size(public_aws_file["path"]) == public_aws_file["file_size"]

    @pytest.mark.skip
    @pytest.mark.requires_hdf
    def test_private_hdf_no_access(self, hdf, private_aws_file):
        """
          Do the reported properties of a dataset match expected values?
        """
        with pytest.raises(PermissionError):
            info = hdf.get_series_info(private_aws_file["path"], columns=private_aws_file["series_name"])

    @pytest.mark.requires_hdf
    @patch.dict('os.environ', {'AWS_ACCESS_KEY_ID': IAM_ACCESS_KEY, 'AWS_SECRET_ACCESS_KEY': IAM_SECRET_KEY})
    def test_private_hdf_correct_access(self, private_aws_file):
        """
          Do the reported properties of a dataset match expected values?
        """

        hdf = HdfStorageAdapter()

        info = hdf.get_series_info(private_aws_file["path"], columns=private_aws_file["series_name"])

        assert info[0]["name"] == private_aws_file["series_name"]
        assert info[0]["length"] == private_aws_file["series_size"]
        assert info[0]["dtype"] == private_aws_file["series_type"]

    @pytest.mark.requires_hdf
    def test_hdf_dataset(self, hdf, public_aws_file):
        """
          Does a specified subset of a dataset match its expected
          index and series values?
        """
        expected = {
            public_aws_file["series_name"]: {
                (8,16): { "1972-01-09": 0.65664,
                          "1972-01-10": 0.65664,
                          "1972-01-11": 0.65664,
                          "1972-01-12": 0.65664,
                          "1972-01-13": 0.65664,
                          "1972-01-14": 0.65664,
                          "1972-01-15": 0.65664,
                          "1972-01-16": 0.65664},

                (12008,12016): { "2004-11-16": 1.2528,
                                 "2004-11-17": 1.2528,
                                 "2004-11-18": 1.2528,
                                 "2004-11-19": 1.2528,
                                 "2004-11-20": 1.2528,
                                 "2004-11-21": 1.2528,
                                 "2004-11-22": 1.2528,
                                 "2004-11-23": 1.2528}
            }
        }
        for series_name, ranges in expected.items():
            for bounds, data in ranges.items():
                df_json = hdf.get_columns_as_dataframe(public_aws_file["path"], columns=[series_name], start=bounds[0], end=bounds[1])
                df = pd.read_json(df_json)
                for ts, val in data.items():
                    assert np.isclose(df[series_name][ts], val)

    @pytest.mark.requires_hdf
    def test_hydra_hdf_size(self, client, public_aws_file):
        """
          Does the reported file size match an expected value when
          accessed via hydra.lib?
        """
        assert client.get_hdf_file_size(public_aws_file["path"]) == public_aws_file["file_size"]

    @pytest.mark.requires_hdf
    def test_hydra_hdf_info(self, client, public_aws_file):
        """
          Do the reported properties of a series match expected values when
          accessed via hydra.lib?
        """
        info = client.get_hdf_group_info(public_aws_file["path"])

        assert info["index"] == {'name': 'timestamp', 'length': 12784, 'dtype': 'datetime64'}
        assert len(info["series"]) == 71
        assert info["series"][0] == {'name': 'BR_Bendera (Ruzizi 0.035)', 'length': 12784, 'dtype': 'float64'}

    @pytest.mark.requires_hdf
    def test_hydra_hdf_dataset(self, client, public_aws_file):
        """
          Does a specified subset of a dataset match its expected
          index and series values when accessed via hydra.lib?
        """
        expected = {
            public_aws_file["series_name"]: {
                (8,16): { "1972-01-09": 0.65664,
                          "1972-01-10": 0.65664,
                          "1972-01-11": 0.65664,
                          "1972-01-12": 0.65664,
                          "1972-01-13": 0.65664,
                          "1972-01-14": 0.65664,
                          "1972-01-15": 0.65664,
                          "1972-01-16": 0.65664},

                (12008,12016): { "2004-11-16": 1.2528,
                                 "2004-11-17": 1.2528,
                                 "2004-11-18": 1.2528,
                                 "2004-11-19": 1.2528,
                                 "2004-11-20": 1.2528,
                                 "2004-11-21": 1.2528,
                                 "2004-11-22": 1.2528,
                                 "2004-11-23": 1.2528}
            }
        }
        for series_name, ranges in expected.items():
            for bounds, series in ranges.items():
                df_json = client.get_hdf_columns_as_dataframe(public_aws_file["path"], groupname=None, columns=[series_name], start=bounds[0], end=bounds[1])
                df = pd.read_json(df_json)
                for ts, val in series.items():
                    assert np.isclose(df[series_name][ts], val)

    @pytest.mark.requires_hdf
    def test_bad_url(self, client, bad_url):
        """
          Does an inaccessible url raise ValueError?
        """
        with pytest.raises(ValueError):
            info = client.get_hdf_series_info(bad_url, "series_name")

    @pytest.mark.requires_hdf
    def test_bad_series_name(self, client, public_aws_file):
        """
          Does a nonexistent dataset name raise ValueError both for
          info and series retrieval?
        """
        with pytest.raises(ValueError):
            info = client.get_hdf_series_info(public_aws_file["path"], "nonexistent_series")

        with pytest.raises(ValueError):
            df_json = client.get_hdf_columns_as_dataframe(public_aws_file["path"], columns=["nonexistent_series"], start=8, end=16)

    @pytest.mark.requires_hdf
    def test_bad_bounds(self, client, public_aws_file):
        """
          Do invalid bounds (start<0, start>end, end<0, end>size) raise ValueError?
        """
        with pytest.raises(ValueError):
            df_json = client.get_hdf_group_as_dataframe(public_aws_file["path"], start=-1, end=16)

        with pytest.raises(ValueError):
            df_json = client.get_hdf_group_as_dataframe(public_aws_file["path"], start=16, end=8)

        with pytest.raises(ValueError):
            df_json = client.get_hdf_group_as_dataframe(public_aws_file["path"], start=8, end=-1)

        with pytest.raises(ValueError):
            df_json = client.get_hdf_group_as_dataframe(public_aws_file["path"], start=8, end=1e72)

    @pytest.mark.requires_hdf
    def test_bad_url_does_not_exist(self, client, bad_url):
        """
          Does data.file_exists_at_url() return False for nonexistent
          files on remote or local filesystem?
        """
        assert not client.file_exists_at_url(bad_url)

    @pytest.mark.requires_hdf
    def test_existing_file_at_url_exists(self, client, public_aws_file):
        """
          Does data.file_exists_at_url() return True for existing
          files on remote or local filesystem?
        """
        assert client.file_exists_at_url(public_aws_file["path"])

    def test_nulladapter_is_null(self):
        """
          If HDF storage were to be disabled in config, does the NullAdapter
          replacement class transparently allow the API to do nothing?
        """
        # Static method
        NullAdapter.method()
        NullAdapter.nested.method()

        # Instance method
        na = NullAdapter()
        na.do.nothing()

        # Is iterable?
        for i in na:
            pass

        # Is subscriptable?
        assert na[2] == None

    @pytest.mark.requires_hdf
    def test_multigroup_series(self, client, multigroup_file):
        """
          Is the correct series returned when requesting a particular
          series from a dataframe in a multigroup file?

          NB rot13 strings
        """
        df = client.get_hdf_columns_as_dataframe(multigroup_file["path"],
                                               groupname="RFJ_Rffrk_erfhygf",
                                               columns=["Jbezvatsbeq Vagnxr.Fhccyl.Nzbhag"],
                                               end=256)
        assert df[:84] == '{"Jbezvatsbeq Vagnxr.Fhccyl.Nzbhag":{"1910-01-01 00:00:00":0.0,"1910-01-02 00:00:00"'

    @pytest.mark.requires_hdf
    def test_hdf_multigroups(self, client, multigroup_file):
        """
          Are the expected groups returned when querying an HDF file for its root groups?

          NB rot13 strings
        """
        groups = client.get_hdf_groups(multigroup_file["path"])
        assert set(groups) == set(multigroup_file["groups"])

    @pytest.mark.requires_hdf
    def test_get_hdf_whole_group_as_dataframe(self, client, multigroup_file):
        """
          Does retrieving a selection of rows from a whole Group
          return the correct section?

          NB rot13 strings
        """
        df_json = client.get_hdf_group_as_dataframe(multigroup_file["path"],
                                                  groupname="RFJ_Rffrk_erfhygf",
                                                  start=4, end=8)
        assert df_json[:114] == '{"Ynatunz Vagnxr.Fhccyl.Nzbhag":{"1910-01-05 00:00:00":40.0,'\
                                '"1910-01-06 00:00:00":40.0,"1910-01-07 00:00:00":40.0,'

    @pytest.mark.requires_hdf
    def test_hdf_group_columns(self, client, multigroup_file):
        """
          Are the correct columns in the correct order returned from a group dataframe?

          NB rot13 strings
        """
        columns = client.get_hdf_group_columns(multigroup_file["path"], groupname="RFJ_Rffrk_erfhygf")
        assert len(columns) == 109
        assert columns[2] == 'Qraire Vagnxr.Fhccyl.Nzbhag'
        assert 'Unaavatsvryq Erfreibve.Fgbentr.Pnyphyngrq (%)' in columns

    @pytest.mark.requires_hdf
    @pytest.mark.parametrize("pathname, path", {**bad_paths, **{"symlink_path": symlink_path}}.items())
    def test_purge_file(self, client, hdf, pathname, path, request):
        """
          This tests the operation of HdfStorageAdapter.purge_local_file()
          This is an inherently dangerous function which allows a file specified
          as an argument to be deleted from the local filesystem.
          The precautions taken against abuse of this are described in the function's
          docstring.
          Success in these tests occurs when an exception is raised as a result of a
          invalid argument.
        """
        fsp = hdf.filestore_path
        try:
            # The symlink_path is just a ref to the fixture func
            # so make pytest replace the ref with its value
            path = request.getfixturevalue(pathname)
        except:
            pass
        with pytest.raises(ValueError):
            client.purge_local_file(os.path.join(fsp, path))

    @pytest.mark.requires_hdf
    @pytest.mark.parametrize("file, file_info", [("public_aws_file", public_aws_file), ("multigroup_file", multigroup_file)])
    def test_index_info(self, client, file, file_info, request):
        try:
            file_info = request.getfixturevalue(file)
        except:
            pass
        groups = client.get_hdf_groups(file_info["path"])
        ii = client.get_hdf_index_info(file_info["path"], groupname=groups[0])
        for key in ("name", "length", "dtype"):
            assert key in ii
        assert isinstance(ii, dict)

    @pytest.mark.requires_hdf
    @pytest.mark.parametrize("file, file_info", [("public_aws_file", public_aws_file), ("multigroup_file", multigroup_file)])
    def test_index_range(self, client, file, file_info, request):
        try:
            file_info = request.getfixturevalue(file)
        except:
            pass
        groups = client.get_hdf_groups(file_info["path"])
        ir = client.get_hdf_index_range(file_info["path"], groupname=groups[0], start=2, end=8)
        assert len(ir) > 0
        assert isinstance(ir[0], str)


    @pytest.mark.requires_hdf
    def test_non_timeseries_index(self, client, non_timeseries_index_file):
        df_json = client.get_hdf_group_as_dataframe(
                    non_timeseries_index_file["path"],
                    groupname=non_timeseries_index_file["group"])
        dfd = json.loads(df_json)
        df = pd.DataFrame(dfd)
        df.index = pd.Index(df.index, dtype=np.float64).sort_values()
        assert len(df) == 11
        assert np.array_equal(df.index, np.linspace(0, 100, 11))
